@@ -14,16 +14,22 @@ from firebirdsql import (DatabaseError, InternalError, OperationalError,
 )
 from firebirdsql.consts import *
 
-DEFAULT_CHARSET='UTF8'
 PYTHON_MAJOR_VER = sys.version_info[0]
 
+def bi(b):
+    "byte to integer"
+    if PYTHON_MAJOR_VER==3:
+        return b
+    return ord(b)
+
 def bs(byte_array):
+    "bytes to string"
     if PYTHON_MAJOR_VER==3:
         return bytes(byte_array)
     return ''.join([chr(c) for c in byte_array])
 
 DEBUG = False
-__version__ = '0.5.0'
+__version__ = '0.6.0'
 apilevel = '2.0'
 threadsafety = 1
 paramstyle = 'qmark'
@@ -38,11 +44,24 @@ transaction_parameter_block = [
     bs([isc_tpb_version3, isc_tpb_write, isc_tpb_wait, isc_tpb_concurrency]),
     # ISOLATION_LEVEL_SERIALIZABLE
     bs([isc_tpb_version3, isc_tpb_write, isc_tpb_wait, isc_tpb_consistency]),
+    # ISOLATION_LEVEL_READ_COMMITED_READ_ONLY
+    bs([isc_tpb_version3, isc_tpb_read, isc_tpb_wait, isc_tpb_read_committed, isc_tpb_no_rec_version]),
 ]
 
-INFO_SQL_STMT_TYPE = bs([0x15])
-INFO_SQL_SQLDA_START = bs([0x14,0x02])
-INFO_SQL_SELECT_DESCRIBE_VARS = bs([0x04,0x07,0x09,0x0b,0x0c,0x0d,0x0e,0x0f,0x10,0x11,0x12,0x13,0x08])
+INFO_SQL_SELECT_DESCRIBE_VARS = bs([
+    isc_info_sql_select,
+    isc_info_sql_describe_vars,
+    isc_info_sql_sqlda_seq,
+    isc_info_sql_type,
+    isc_info_sql_sub_type,
+    isc_info_sql_scale,
+    isc_info_sql_length,
+    isc_info_sql_null_ind,
+    isc_info_sql_field,
+    isc_info_sql_relation,
+    isc_info_sql_owner,
+    isc_info_sql_alias,
+    isc_info_sql_describe_end])
 
 def Date(year, month, day):
     return datetime.date(year, month, day)
@@ -312,160 +331,210 @@ def calc_blr(xsqlda):
     # x.sqlscale value shoud be negative, so b convert to range(0, 256)
     return bs([256 + b if b < 0 else b for b in blr])
 
-
-
-class cursor:
-    def _parse_select_items(self, buf):
-        index = 0
-        i = 0
-        if PYTHON_MAJOR_VER==3:
-            item = buf[i]
+def parse_select_items(buf, xsqlda, connection):
+    index = 0
+    i = 0
+    item = bi(buf[i])
+    while item != isc_info_end:
+        if item == isc_info_sql_sqlda_seq:
+            l = bytes_to_int(buf[i+1:i+3])
+            index = bytes_to_int(buf[i+3:i+3+l])
+            xsqlda[index-1] = XSQLVAR(connection.bytes_to_str)
+            i = i + 3 + l
+        elif item == isc_info_sql_type:
+            l = bytes_to_int(buf[i+1:i+3])
+            xsqlda[index-1].sqltype = bytes_to_int(buf[i+3:i+3+l]) & ~ 1
+            i = i + 3 + l
+        elif item == isc_info_sql_sub_type:
+            l = bytes_to_int(buf[i+1:i+3])
+            xsqlda[index-1].sqlsubtype = bytes_to_int(buf[i+3:i+3+l])
+            i = i + 3 + l
+        elif item == isc_info_sql_scale:
+            l = bytes_to_int(buf[i+1:i+3])
+            xsqlda[index-1].sqlscale = bytes_to_int(buf[i+3:i+3+l])
+            i = i + 3 + l
+        elif item == isc_info_sql_length:
+            l = bytes_to_int(buf[i+1:i+3])
+            xsqlda[index-1].sqllen = bytes_to_int(buf[i+3:i+3+l])
+            i = i + 3 + l
+        elif item == isc_info_sql_null_ind:
+            l = bytes_to_int(buf[i+1:i+3])
+            xsqlda[index-1].null_ok = bytes_to_int(buf[i+3:i+3+l])
+            i = i + 3 + l
+        elif item == isc_info_sql_field:
+            l = bytes_to_int(buf[i+1:i+3])
+            xsqlda[index-1].fieldname = connection.bytes_to_str(buf[i+3:i+3+l])
+            i = i + 3 + l
+        elif item == isc_info_sql_relation:
+            l = bytes_to_int(buf[i+1:i+3])
+            xsqlda[index-1].relname = connection.bytes_to_str(buf[i+3:i+3+l])
+            i = i + 3 + l
+        elif item == isc_info_sql_owner:
+            l = bytes_to_int(buf[i+1:i+3])
+            xsqlda[index-1].ownname = connection.bytes_to_str(buf[i+3:i+3+l])
+            i = i + 3 + l
+        elif item == isc_info_sql_alias:
+            l = bytes_to_int(buf[i+1:i+3])
+            xsqlda[index-1].aliasname = connection.bytes_to_str(buf[i+3:i+3+l])
+            i = i + 3 + l
+        elif item == isc_info_truncated:
+            return index    # return next index
+        elif item == isc_info_sql_describe_end:
+            i = i + 1
         else:
-            item = ord(buf[i])
-        while item != isc_info_end:
-            if item == isc_info_sql_sqlda_seq:
-                l = bytes_to_int(buf[i+1:i+3])
-                index = bytes_to_int(buf[i+3:i+3+l])
-                self._xsqlda[index-1] = XSQLVAR(self.connection.bytes_to_str)
-                i = i + 3 + l
-            elif item == isc_info_sql_type:
-                l = bytes_to_int(buf[i+1:i+3])
-                self._xsqlda[index-1].sqltype = \
-                                        bytes_to_int(buf[i+3:i+3+l]) & ~ 1
-                i = i + 3 + l
-            elif item == isc_info_sql_sub_type:
-                l = bytes_to_int(buf[i+1:i+3])
-                self._xsqlda[index-1].sqlsubtype = bytes_to_int(buf[i+3:i+3+l])
-                i = i + 3 + l
-            elif item == isc_info_sql_scale:
-                l = bytes_to_int(buf[i+1:i+3])
-                self._xsqlda[index-1].sqlscale = bytes_to_int(buf[i+3:i+3+l])
-                i = i + 3 + l
-            elif item == isc_info_sql_length:
-                l = bytes_to_int(buf[i+1:i+3])
-                self._xsqlda[index-1].sqllen = bytes_to_int(buf[i+3:i+3+l])
-                i = i + 3 + l
-            elif item == isc_info_sql_null_ind:
-                l = bytes_to_int(buf[i+1:i+3])
-                self._xsqlda[index-1].null_ok = bytes_to_int(buf[i+3:i+3+l])
-                i = i + 3 + l
-            elif item == isc_info_sql_field:
-                l = bytes_to_int(buf[i+1:i+3])
-                self._xsqlda[index-1].fieldname = \
-                        self.connection.bytes_to_str(buf[i + 3: i + 3 + l])
-                i = i + 3 + l
-            elif item == isc_info_sql_relation:
-                l = bytes_to_int(buf[i+1:i+3])
-                self._xsqlda[index-1].relname = \
-                        self.connection.bytes_to_str(buf[i + 3: i + 3 + l])
-                i = i + 3 + l
-            elif item == isc_info_sql_owner:
-                l = bytes_to_int(buf[i+1:i+3])
-                self._xsqlda[index-1].ownname = \
-                        self.connection.bytes_to_str(buf[i + 3: i + 3 + l])
-                i = i + 3 + l
-            elif item == isc_info_sql_alias:
-                l = bytes_to_int(buf[i+1:i+3])
-                self._xsqlda[index-1].aliasname = \
-                        self.connection.bytes_to_str(buf[i + 3: i + 3 + l])
-                i = i + 3 + l
-            elif item == isc_info_truncated:
-                return index    # return next index
-            elif item == isc_info_sql_describe_end:
-                i = i + 1
-            else:
-                print('\t', item, 'Invalid item [%02x] ! i=%d' % (buf[i], i))
-                i = i + 1
-            if PYTHON_MAJOR_VER==3:
-                item = buf[i]
-            else:
-                item = ord(buf[i])
-        return -1   # no more info
+            print('\t', item, 'Invalid item [%02x] ! i=%d' % (buf[i], i))
+            i = i + 1
+        item = bi(buf[i])
+    return -1   # no more info
 
-    def __init__(self, conn):
-        if not hasattr(conn, "db_handle"):
-            raise InternalError()
-        self.connection = conn
-        if not hasattr(self.connection, "trans_handle"):
-            self.connection.begin()
-        self.connection._op_allocate_statement()
-        (h, oid, buf) = self.connection._op_response()
+def parse_xsqlda(buf, connection, stmt_handle):
+    assert buf[:3] == bs([0x15,0x04,0x00]) # isc_info_sql_stmt_type
+    stmt_type = bytes_to_int(buf[3:7])
+    if stmt_type != isc_info_sql_stmt_select:
+        return []
+
+    assert buf[7:9] == bs([0x04,0x07])
+    l = bytes_to_int(buf[9:11])
+    col_len = bytes_to_int(buf[11:11+l])
+    xsqlda = [None] * col_len
+    next_index = parse_select_items(buf[11+l:], xsqlda, connection)
+    while next_index > 0:   # more describe vars
+        self.connection._op_info_sql(stmt_handle,
+                    bs([isc_info_sql_sqlda_start, 2])
+                        + int_to_bytes(next_index, 2)
+                        + INFO_SQL_SELECT_DESCRIBE_VARS)
+        (h, oid, buf) = connection._op_response()
+        assert buf[:2] == bs([0x04,0x07])
+        l = bytes_to_int(buf[2:4])
+        assert bytes_to_int(buf[4:4+l]) == col_len
+        next_index = parse_select_items(buf[4+l:], xsqlda, connection)
+    return xsqlda
+
+class PreparedStatement:
+    def __init__(self, cur, sql):
+        self.cur = cur
+        self.sql = sql
+        transaction = self.cur.transaction
+        connection = transaction.connection
+
+        connection._op_allocate_statement(transaction)
+        (h, oid, buf) = connection._op_response()
+        self.stmt_handle = h
+
+        connection._op_prepare_statement(
+                self.stmt_handle, transaction.trans_handle, sql, 
+                option_items=bs([isc_info_sql_get_plan]))
+        (h, oid, buf) = connection._op_response()
+
+        i = 0
+        if bi(buf[i]) == isc_info_sql_get_plan:
+            l = bytes_to_int(buf[i+1:i+3])
+            self.plan = connection.bytes_to_str(buf[i+3:i+3+l])
+            i += 3 + l
+
+        assert buf[i:i+3] == bs([0x15,0x04,0x00]) # isc_info_sql_stmt_type (4 bytes)
+        self.statement_type = bytes_to_int(buf[i+3:i+7])
+        self._xsqlda = parse_xsqlda(buf[i:], connection, self.stmt_handle)
+
+        # TODO: implement later
+        self.n_input_params = None
+
+    def __getattr__(self, attrname):
+        if attrname == 'description':
+            if len(self._xsqlda) == 0:
+                return None
+            r = []
+            for x in self._xsqlda:
+                r.append((x.aliasname, x.sqltype, None, x.io_length(), None, 
+                        x.sqlscale, True if x.null_ok else False))
+            return r
+        elif attrname == 'n_output_params':
+            return len(self._xsqlda)
+        raise AttributeError
+
+class Cursor:
+    def __init__(self, trans):
+        self._transaction = trans
+        self.transaction.connection._op_allocate_statement(self.transaction)
+        (h, oid, buf) = self.transaction.connection._op_response()
         self.stmt_handle = h
         self.arraysize = 1
 
+    @property
+    def transaction(self):
+        return self._transaction
+
     def callproc(self, procname, *params):
         raise NotSupportedError()
 
-    def _execute(self, query, params):
-        self.connection._op_prepare_statement(self.stmt_handle, query)
-        (h, oid, buf) = self.connection._op_response()
-
+    def _convert_params(self, params):
         cooked_params = []
         for param in params:        # Convert str/bytes parameter to blob id
             if type(param) == str:
-                param = self.connection.str_to_bytes(param)
-            else:
-                cooked_params.append(param)
-                continue
-            self.connection._op_create_blob2()
-            (blob_handle, blob_id, buf2) = self.connection._op_response()
-            seg_size = self.connection.buffer_length
+                param = self.transaction.connection.str_to_bytes(param)
+            cooked_params.append(param)
+            continue
+            self.transaction.connection._op_create_blob2()
+            (blob_handle, blob_id, buf2) = self.transaction.connection._op_response()
+            seg_size = self.transaction.connection.buffer_length
             (seg, remains) = param[:seg_size], param[seg_size:]
             while seg:
-                self.connection._op_batch_segments(blob_handle, seg)
-                (h3, oid3, buf3) = self.connection._op_response()
+                self.transaction.connection._op_batch_segments(blob_handle, seg)
+                (h3, oid3, buf3) = self.transaction.connection._op_response()
                 (seg, remains) = remains[:seg_size], remains[seg_size:]
-            self.connection._op_close_blob(blob_handle)
-            (h4, oid4, buf4) = self.connection._op_response()
+            self.transaction.connection._op_close_blob(blob_handle)
+            (h4, oid4, buf4) = self.transaction.connection._op_response()
             assert blob_id == oid4
             cooked_params.append(blob_id)
+        return cooked_params
 
-        assert buf[:3] == bs([0x15,0x04,0x00]) # isc_info_sql_stmt_type (4 bytes)
-        stmt_type = bytes_to_int(buf[3:7])
-        if stmt_type == isc_info_sql_stmt_select:
-            assert buf[7:9] == bs([0x04,0x07])
-            l = bytes_to_int(buf[9:11])
-            col_len = bytes_to_int(buf[11:11+l])
-            self._xsqlda = [None] * col_len
-            next_index = self._parse_select_items(buf[11+l:])
-            while next_index > 0:   # more describe vars
-                self.connection._op_info_sql(self.stmt_handle,
-                            INFO_SQL_SQLDA_START + int_to_bytes(next_index, 2) 
-                            + INFO_SQL_SELECT_DESCRIBE_VARS)
-                (h, oid, buf) = self.connection._op_response()
-                assert buf[:2] == bs([0x04,0x07])
-                l = bytes_to_int(buf[2:4])
-                assert bytes_to_int(buf[4:4+l]) == col_len
-                next_index = self._parse_select_items(buf[4+l:])
-
-            self.connection._op_execute(self.stmt_handle, cooked_params)
-            (h, oid, buf) = self.connection._op_response()
+    def _execute(self, query, params):
+        cooked_params = self._convert_params(params)
 
+        if isinstance(query, PreparedStatement):
+            stmt_handle = query.stmt_handle
+            stmt_type = query.statement_type
+            self._xsqlda = query._xsqlda
         else:
-            self.connection._op_execute(self.stmt_handle, cooked_params)
-            try:
-                (h, oid, buf) = self.connection._op_response()
-            except OperationalError:
-                e = sys.exc_info()[1]
-                if 335544665 in e.gds_codes:
-                    raise IntegrityError(e.message, e.gds_codes, e.sql_code)
-        return stmt_type
+            stmt_handle = self.stmt_handle
+            self.transaction.connection._op_prepare_statement(stmt_handle, 
+                                        self.transaction.trans_handle, query)
+            (h, oid, buf) = self.transaction.connection._op_response()
+            assert buf[:3] == bs([0x15,0x04,0x00]) # isc_info_sql_stmt_type
+            stmt_type = bytes_to_int(buf[3:7])
+            self._xsqlda = parse_xsqlda(buf, self.transaction.connection, 
+                                                                stmt_handle)
+
+        self.transaction.connection._op_execute(stmt_handle, 
+                                self.transaction.trans_handle, cooked_params)
+        try:
+            (h, oid, buf) = self.transaction.connection._op_response()
+        except OperationalError:
+            e = sys.exc_info()[1]
+            if 335544665 in e.gds_codes:
+                raise IntegrityError(e.message, e.gds_codes, e.sql_code)
+        return stmt_type, stmt_handle
+
+    def prep(self, query):
+        prepared_statement = PreparedStatement(self, query)
+        return prepared_statement
 
     def execute(self, query, params = []):
-        self._stmt_type = self._execute(query, params)
-        if self._stmt_type == isc_info_sql_stmt_select:
-            self._fetch_records = self._fetch_generator()
+        stmt_type, stmt_handle = self._execute(query, params)
+        if stmt_type == isc_info_sql_stmt_select:
+            self._fetch_records = self._fetch_generator(stmt_handle)
 
     def executemany(self, query, seq_of_params):
         for params in seq_of_params:
             self.execute(query, params)
 
-    def _fetch_generator(self):
+    def _fetch_generator(self, stmt_handle):
+        connection = self.transaction.connection
         more_data = True
         while more_data:
-            self.connection._op_fetch(self.stmt_handle, calc_blr(self._xsqlda))
-            (rows, more_data) = self.connection._op_fetch_response(
-                                            self.stmt_handle, self._xsqlda)
+            connection._op_fetch(stmt_handle, calc_blr(self._xsqlda))
+            (rows, more_data) = connection._op_fetch_response(
+                                            stmt_handle, self._xsqlda)
             for r in rows:
                 # Convert BLOB handle to data    
                 for i in range(len(self._xsqlda)):    
@@ -473,27 +542,27 @@ class cursor:
                     if x.sqltype == SQL_TYPE_BLOB:    
                         if not r[i]:
                             continue
-                        self.connection._op_open_blob(r[i])
-                        (h, oid, buf) = self.connection._op_response()
+                        connection._op_open_blob(r[i])
+                        (h, oid, buf) = connection._op_response()
                         v = bs([])
                         n = 1   # 1:mora data 2:no more data
                         while n == 1:
-                            self.connection._op_get_segment(h)
-                            (n, oid, buf) = self.connection._op_response()
+                            connection._op_get_segment(h)
+                            (n, oid, buf) = connection._op_response()
                             while buf:
                                 ln = bytes_to_int(buf[:2])
                                 v += buf[2:ln+2]
                                 buf = buf[ln+2:]
-                        self.connection._op_close_blob(h)
-                        (h, oid, buf) = self.connection._op_response()
+                        connection._op_close_blob(h)
+                        (h, oid, buf) = connection._op_response()
                         r[i] = v
                 yield r
 
         # recreate stmt_handle
-        self.connection._op_free_statement(self.stmt_handle, 2) # DSQL_drop
-        (h, oid, buf) = self.connection._op_response()
-        self.connection._op_allocate_statement()
-        (h, oid, buf) = self.connection._op_response()
+        connection._op_free_statement(stmt_handle, 2) # DSQL_drop
+        (h, oid, buf) = connection._op_response()
+        connection._op_allocate_statement(self.transaction)
+        (h, oid, buf) = connection._op_response()
         self.stmt_handle = h
 
         raise StopIteration()
@@ -526,8 +595,8 @@ class cursor:
     def close(self):
         if not hasattr(self, "stmt_handle"):
             return
-        self.connection._op_free_statement(self.stmt_handle, 2)   # DSQL_drop
-        (h, oid, buf) = self.connection._op_response()
+        self.transaction.connection._op_free_statement(self.stmt_handle, 2)   # DSQL_drop
+        (h, oid, buf) = self.transaction.connection._op_response()
         delattr(self, "stmt_handle")
 
     def nextset():
@@ -550,11 +619,7 @@ class cursor:
             return -1
         raise AttributeError
 
-    def __del__(self):
-        if hasattr(self, "stmt_handle"):
-            self.close()
-
-class BaseConnect:
+class Connection:
     buffer_length = 1024
     op_connect = 1
     op_accept = 3
@@ -638,6 +703,13 @@ class BaseConnect:
         return s.encode(self.charset_map.get(self.charset, self.charset))
 
     def bytes_to_str(self, b):
+        "convert bytes array to raw string"
+        if PYTHON_MAJOR_VER == 3:
+            return b.decode(self.charset_map.get(self.charset, self.charset))
+        return b
+
+    def bytes_to_ustr(self, b):
+        "convert bytes array to unicode string"
         return b.decode(self.charset_map.get(self.charset, self.charset))
 
     def uid(self):
@@ -771,10 +843,10 @@ class BaseConnect:
     def _op_service_attach(self):
         dpb = bs([2,2])
         s = self.str_to_bytes(self.user)
-        dpb += bs([28, len(s)]) + s
+        dpb += bs([isc_spb_user_name, len(s)]) + s
         s = self.str_to_bytes(self.password)
-        dpb += bs([29, len(s)]) + s
-        dpb += bs([0x3a,0x04,0x78,0x0a,0x00,0x00])  # isc_dpb_dummy_packet_interval
+        dpb += bs([isc_spb_password, len(s)]) + s
+        dpb += bs([isc_spb_dummy_packet_interval,0x04,0x78,0x0a,0x00,0x00])
         p = xdrlib.Packer()
         p.pack_int(self.op_service_attach)
         p.pack_int(0)
@@ -786,7 +858,8 @@ class BaseConnect:
     def _op_service_info(self, param, item, buffer_length=512):
         p = xdrlib.Packer()
         p.pack_int(self.op_service_info)
-        p.pack_int(self.svc_handle)
+#        p.pack_int(self.svc_handle)
+        p.pack_int(self.db_handle)
         p.pack_int(0)
         p.pack_bytes(param)
         p.pack_bytes(item)
@@ -828,35 +901,35 @@ class BaseConnect:
         send_channel(self.sock, p.get_buffer())
 
     @wire_operation
-    def _op_commit(self):
+    def _op_commit(self, trans_handle):
         p = xdrlib.Packer()
         p.pack_int(self.op_commit)
-        p.pack_int(self.trans_handle)
+        p.pack_int(trans_handle)
         send_channel(self.sock, p.get_buffer())
 
     @wire_operation
-    def _op_commit_retaining(self):
+    def _op_commit_retaining(self, trans_handle):
         p = xdrlib.Packer()
         p.pack_int(self.op_commit_retaining)
-        p.pack_int(self.trans_handle)
+        p.pack_int(trans_handle)
         send_channel(self.sock, p.get_buffer())
 
     @wire_operation
-    def _op_rollback(self):
+    def _op_rollback(self, trans_handle):
         p = xdrlib.Packer()
         p.pack_int(self.op_rollback)
-        p.pack_int(self.trans_handle)
+        p.pack_int(trans_handle)
         send_channel(self.sock, p.get_buffer())
 
     @wire_operation
-    def _op_rollback_retaining(self):
+    def _op_rollback_retaining(self, trans_handle):
         p = xdrlib.Packer()
         p.pack_int(self.op_rollback_retaining)
-        p.pack_int(self.trans_handle)
+        p.pack_int(trans_handle)
         send_channel(self.sock, p.get_buffer())
 
     @wire_operation
-    def _op_allocate_statement(self):
+    def _op_allocate_statement(self, trans_handle):
         p = xdrlib.Packer()
         p.pack_int(self.op_allocate_statement)
         p.pack_int(self.db_handle)
@@ -871,14 +944,15 @@ class BaseConnect:
         send_channel(self.sock, p.get_buffer())
 
     @wire_operation
-    def _op_prepare_statement(self, stmt_handle, query):
+    def _op_prepare_statement(self, stmt_handle, trans_handle, query, option_items=bs([])):
+        desc_items = option_items + bs([isc_info_sql_stmt_type])+INFO_SQL_SELECT_DESCRIBE_VARS
         p = xdrlib.Packer()
         p.pack_int(self.op_prepare_statement)
-        p.pack_int(self.trans_handle)
+        p.pack_int(trans_handle)
         p.pack_int(stmt_handle)
         p.pack_int(3)   # dialect = 3
         p.pack_string(self.str_to_bytes(query))
-        p.pack_bytes(INFO_SQL_STMT_TYPE + INFO_SQL_SELECT_DESCRIBE_VARS)
+        p.pack_bytes(desc_items)
         p.pack_int(self.buffer_length)
         send_channel(self.sock, p.get_buffer())
 
@@ -893,11 +967,11 @@ class BaseConnect:
         send_channel(self.sock, p.get_buffer())
 
     @wire_operation
-    def _op_execute(self, stmt_handle, params):
+    def _op_execute(self, stmt_handle, trans_handle, params):
         p = xdrlib.Packer()
         p.pack_int(self.op_execute)
         p.pack_int(stmt_handle)
-        p.pack_int(self.trans_handle)
+        p.pack_int(trans_handle)
 
         if len(params) == 0:
             p.pack_bytes(bs([]))
@@ -1047,48 +1121,37 @@ class BaseConnect:
         return (h, oid, buf)
 
     def cursor(self):
-        c = cursor(self)
-        self.cursor_set.add(c)
+        if self.main_transaction is None:
+            self.begin()
+        c = Cursor(self.main_transaction)
         return c
 
     def begin(self):
         if not hasattr(self, "db_handle"):
             raise InternalError
-        self._op_transaction(transaction_parameter_block[self.isolation_level])
-        (h, oid, buf) = self._op_response()
-        self.trans_handle = h
+        trans = Transaction(self)
+        trans.begin()
+
+    @property
+    def transactions(self):
+        return self._transactions
+
+    @property
+    def main_transaction(self):
+        if len(self._transactions):
+            return self._transactions[0]
+        return None
     
     def commit(self, retaining=False):
-        if retaining:
-            self._op_commit_retaining()
-            (h, oid, buf) = self._op_response()
-        else:
-            self._op_commit()
-            (h, oid, buf) = self._op_response()
-            delattr(self, "trans_handle")
+        self.main_transaction.commit(retaining=retaining)
 
     def rollback(self, retaining=False):
-        if retaining:
-            self._op_rollback_retaining()
-            (h, oid, buf) = self._op_response()
-        else:
-            self._op_rollback()
-            (h, oid, buf) = self._op_response()
-            delattr(self, "trans_handle")
-
-    def close(self):
-        if not hasattr(self, "db_handle"):
-            return
-        for c in self.cursor_set:
-            c.close()
-        self._op_rollback()
-        (h, oid, buf) = self._op_response()
-        self._op_detach()
-        (h, oid, buf) = self._op_response()
-        delattr(self, "db_handle")
+        self.main_transaction.rollback(retaining=retaining)
 
     def __init__(self, dsn=None, user=None, password=None, host=None,
-            database=None, charset=DEFAULT_CHARSET, port=3050):
+                    database=None, charset=DEFAULT_CHARSET, port=3050, 
+                    page_size=None,
+                    is_services = False):
         if dsn:
             i = dsn.find(':')
             if i < 0:
@@ -1111,199 +1174,188 @@ class BaseConnect:
         self.user = user
         self.password = password
         self.charset = charset
-        self.cursor_set = set()
+        self._transactions = []
         self.isolation_level = ISOLATION_LEVEL_READ_COMMITED
 
         self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
         self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
         self.sock.connect((self.hostname, self.port))
 
+        self.page_size = page_size
+        self.is_services = is_services
+        self._op_connect()
+        self._op_accept()
+        if self.page_size:                      # create database
+            self._op_create(self.page_size)
+        elif self.is_services:                  # service api
+            self._op_service_attach()
+        else:                                   # connect
+            self._op_attach()
+        (h, oid, buf) = self._op_response()
+        self.db_handle = h
+
     def set_isolation_level(self, isolation_level):
         self.isolation_level = isolation_level
 
-    def info_database(self, info_names):
-        if info_names[-1] != isc_info_end:
-            info_names.append(isc_info_end)
-        b = bs(info_names)
-        self._op_info_database(b)
+    def _db_info(self, info_requests):
+        if info_requests[-1] == isc_info_end:
+            self._op_info_database(bs(info_requests))
+        else:
+            self._op_info_database(
+                        bs(info_requests+type(info_requests)([isc_info_end])))
         (h, oid, buf) = self._op_response()
         i = 0
+        i_request = 0
         r = []
         while i < len(buf):
-            if PYTHON_MAJOR_VER==3:
-                s = buf[i]
-            else:
-                s = ord(buf[i])
-            if s == isc_info_end:
+            req = bi(buf[i])
+            if req == isc_info_end:
                 break
-            l = bytes_to_int(buf[i+1:i+3])
-            r.append(buf[i+3:i+3+l])
-            i = i + 3 + l
+            assert req == info_requests[i_request]
+            if req == isc_info_user_names:
+                user_names = []
+                while req == isc_info_user_names:
+                    l = bytes_to_int(buf[i+1:i+3])
+                    user_names.append(buf[i+3:i+3+l])
+                    i = i + 3 + l
+                    req = bi(buf[i])
+                r.append(user_names)
+            else:
+                l = bytes_to_int(buf[i+1:i+3])
+                r.append(buf[i+3:i+3+l])
+                i = i + 3 + l
+            i_request += 1
         return r
 
-    def __del__(self):
-        if hasattr(self, "db_handle"):
-            self.close()
+    def _db_info_convert_type(self, info_request, v):
+        REQ_INT = [
+            isc_info_allocation, isc_info_no_reserve, isc_info_db_sql_dialect,
+            isc_info_ods_minor_version, isc_info_ods_version, 
+            isc_info_page_size, isc_info_current_memory, isc_info_forced_writes,
+            isc_info_max_memory, isc_info_num_buffers, isc_info_sweep_interval,
+            isc_info_limbo, isc_info_attachment_id, isc_info_fetches, 
+            isc_info_marks, isc_info_reads, isc_info_writes, 
+            isc_info_set_page_buffers, isc_info_db_read_only,
+            isc_info_db_size_in_pages, isc_info_page_errors, 
+            isc_info_record_errors, isc_info_bpage_errors, 
+            isc_info_dpage_errors, isc_info_ipage_errors,
+            isc_info_ppage_errors, isc_info_tpage_errors,
+        ]
+        REQ_COUNT = [
+            isc_info_backout_count, isc_info_delete_count, 
+            isc_info_expunge_count, isc_info_insert_count, isc_info_purge_count,
+            isc_info_read_idx_count, isc_info_read_seq_count, 
+            isc_info_update_count
+        ]
 
-class connect(BaseConnect):
-    def __init__(self, dsn=None, user=None, password=None, host=None,
-            database=None, charset=DEFAULT_CHARSET, port=3050):
-        BaseConnect.__init__(self, dsn=dsn, user=user, password=password,
-                    host=host, database=database, charset=charset, port=port)
-        self._op_connect()
-        self._op_accept()
-        self._op_attach()
-        (h, oid, buf) = self._op_response()
-        self.db_handle = h
+        if info_request in (isc_info_base_level, ):
+            # IB6 API guide p52
+            return bi(v[1])
+        elif info_request in (isc_info_db_id, ):
+            # IB6 API guide p52
+            conn_code = bi(v[0])
+            len1 = bi(v[1])
+            filename = self.bytes_to_str(v[2:2+len1])
+            len2 = bi(v[2+len1])
+            sitename = self.bytes_to_str(v[3+len1:3+len1+len2])
+            return (conn_code, filename, sitename)
+        elif info_request in (isc_info_implementation, ):
+            return (bi(v[1]), bi(v[2]))
+        elif info_request in (isc_info_version, isc_info_firebird_version):
+            # IB6 API guide p53
+            return self.bytes_to_str(v[2:2+bi(v[1])])
+        elif info_request in (isc_info_user_names, ):
+            # IB6 API guide p54
+            user_names = []
+            for u in v:
+                user_names.append(self.bytes_to_str(u[1:]))
+            return user_names
+        elif info_request in REQ_INT:
+            return bytes_to_int(v)
+        elif info_request in REQ_COUNT:
+            counts = {}
+            i = 0
+            while i < len(v):
+                counts[bytes_to_int(v[i:i+2])] = bytes_to_int(v[i+2:i+6])
+                i += 6
+            return counts
+        else:
+            return v
 
-
-class create_database(BaseConnect):
-    def __init__(self, dsn=None, user=None, password=None, host=None,
-            database=None, charset=DEFAULT_CHARSET, port=3050, page_size=4096):
-        self.page_size = page_size
-        BaseConnect.__init__(self, dsn=dsn, user=user, password=password,
-                    host=host, database=database, charset=charset, port=port)
-        self._op_connect()
-        self._op_accept()
-        self._op_create(self.page_size)
-        (h, oid, buf) = self._op_response()
-        self.db_handle = h
-
-class service_mgr(BaseConnect):
-    def __init__(self, dsn=None, user=None, password=None, host=None,
-            database=None, charset=DEFAULT_CHARSET, port=3050):
-        BaseConnect.__init__(self, dsn=dsn, user=user, password=password,
-                    host=host, database=database, charset=charset, port=port)
-        self._op_connect()
-        self._op_accept()
-        self._op_service_attach()
-        (h, oid, buf) = self._op_response()
-        self.db_handle = h
-
-    def backup_database(self, backup_filename, callback=None):
-        spb = bs([isc_action_svc_backup])
-        s = self.str_to_bytes(self.filename)
-        spb += bs([isc_spb_dbname]) + int_to_bytes(len(s), 2) + s
-        s = self.str_to_bytes(backup_filename)
-        spb += bs([isc_spb_bkp_file]) + int_to_bytes(len(s), 2) + s
-        if callback:
-            spb += bs([isc_spb_verbose])
-        self._op_service_start(spb)
-        (h, oid, buf) = self._op_response()
-        self.svc_handle = h
-        while True:
-            self._op_service_info(bs([0x02]), bs([0x3e]))
-            (h, oid, buf) = self._op_response()
-            if buf[:4] == bs([0x3e,0x00,0x00,0x01]):
-                break
-            ln = bytes_to_int(buf[1:2])
-            if callback:
-                callback(self.bytes_to_str(buf[3:3+ln]))
-
-    def restore_database(self, restore_filename, callback=None):
-        spb = bs([isc_action_svc_restore])
-        s = self.str_to_bytes(restore_filename)
-        spb += bs([isc_spb_bkp_file]) + int_to_bytes(len(s), 2) + s
-        s = self.str_to_bytes(self.filename)
-        spb += bs([isc_spb_dbname]) + int_to_bytes(len(s), 2) + s
-        if callback:
-            spb += bs([isc_spb_verbose])
-        spb += bs([isc_spb_res_buffers,0x00,0x08,0x00,0x00,isc_spb_res_page_size,0x00,0x10,0x00,0x00,isc_spb_options,0x00,0x30,0x00,0x00])
-        self._op_service_start(spb)
-        (h, oid, buf) = self._op_response()
-        self.svc_handle = h
-        while True:
-            self._op_service_info(bs([0x02]), bs([0x3e]))
-            (h, oid, buf) = self._op_response()
-            if buf[:4] == bs([0x3e,0x00,0x00,0x01]):
-                break
-            ln = bytes_to_int(buf[1:2])
-            if callback:
-                callback(self.bytes_to_str(buf[3:3+ln]))
-
-    def trace_start(self, name=None, cfg=None, callback=None):
-        spb = bs([isc_action_svc_trace_start])
-        if name:
-            s = self.str_to_bytes(name)
-            spb += bs([isc_spb_trc_name]) + int_to_bytes(len(s), 2) + s
-        if cfg:
-            s = self.str_to_bytes(cfg)
-            spb += bs([isc_spb_trc_cfg]) + int_to_bytes(len(s), 2) + s
-        self._op_service_start(spb)
-        (h, oid, buf) = self._op_response()
-        self.svc_handle = h
-        while True:
-            self._op_service_info(bs([0x02]), bs([0x3e]))
-            (h, oid, buf) = self._op_response()
-            if buf[:4] == bs([0x3e,0x00,0x00,0x01]):
-                break
-            ln = bytes_to_int(buf[1:2])
-            if callback:
-                callback(self.bytes_to_str(buf[3:3+ln]))
-
-    def trace_stop(self, id, callback=None):
-        id = int(id)
-        spb = bs([isc_action_svc_trace_stop])
-        spb += bs([isc_spb_trc_id]) + int_to_bytes(id, 4)
-        self._op_service_start(spb)
-        (h, oid, buf) = self._op_response()
-        self.svc_handle = h
-
-        self._op_service_info(bs([0x02]), bs([0x3e]))
-        (h, oid, buf) = self._op_response()
-        ln = bytes_to_int(buf[1:2])
-        if callback:
-            callback(self.bytes_to_str(buf[3:3+ln]))
-
-    def trace_suspend(self, id, callback=None):
-        id = int(id)
-        spb = bs([isc_action_svc_trace_suspend])
-        spb += bs([isc_spb_trc_id]) + int_to_bytes(id, 4)
-        self._op_service_start(spb)
-        (h, oid, buf) = self._op_response()
-        self.svc_handle = h
-
-        self._op_service_info(bs([0x02]), bs([0x3e]))
-        (h, oid, buf) = self._op_response()
-        ln = bytes_to_int(buf[1:2])
-        if callback:
-            callback(self.bytes_to_str(buf[3:3+ln]))
-
-    def trace_resume(self, id, callback=None):
-        id = int(id)
-        spb = bs([isc_action_svc_trace_resume])
-        spb += bs([isc_spb_trc_id]) + int_to_bytes(id, 4)
-        self._op_service_start(spb)
-        (h, oid, buf) = self._op_response()
-        self.svc_handle = h
-
-        self._op_service_info(bs([0x02]), bs([0x3e]))
-        (h, oid, buf) = self._op_response()
-        ln = bytes_to_int(buf[1:2])
-        if callback:
-            callback(self.bytes_to_str(buf[3:3+ln]))
-
-    def trace_list(self, callback=None):
-        spb = bs([isc_action_svc_trace_list])
-        self._op_service_start(spb)
-        (h, oid, buf) = self._op_response()
-        self.svc_handle = h
-        while True:
-            self._op_service_info(bs([0x02]), bs([0x3e]))
-            (h, oid, buf) = self._op_response()
-            if buf[:4] == bs([0x3e,0x00,0x00,0x01]):
-                break
-            ln = bytes_to_int(buf[1:2])
-            if callback:
-                callback(self.bytes_to_str(buf[3:3+ln]))
+    def db_info(self, info_requests):
+        if type(info_requests) == int:  # singleton
+            r = self._db_info([info_requests])
+            return self._db_info_convert_type(info_requests, r[0])
+        else:
+            results = {}
+            rs = self._db_info(info_requests)
+            for i in range(len(info_requests)):
+                results[info_requests[i]] =  self._db_info_convert_type(
+                                                    info_requests[i], rs[i])
+            return results
 
     def close(self):
         if not hasattr(self, "db_handle"):
             return
-        self._op_service_detach()
+        for trans in self._transactions:
+            trans.close()
+        if self.is_services:
+            self._op_service_detach()
+        else:
+            self._op_detach()
         (h, oid, buf) = self._op_response()
         delattr(self, "db_handle")
 
+
     def __del__(self):
         if hasattr(self, "db_handle"):
             self.close()
+
+class Transaction:
+    def __init__(self, connection, tpb=None):
+        self._connection = connection
+
+    def begin(self):
+        self.close()
+        self.connection._op_transaction(
+                transaction_parameter_block[self.connection.isolation_level])
+        (h, oid, buf) = self.connection._op_response()
+        self.trans_handle = h
+        self.connection._transactions.append(self)
+
+    def commit(self, retaining=False):
+        if retaining:
+            self.connection._op_commit_retaining(self.trans_handle)
+            (h, oid, buf) = self.connection._op_response()
+        else:
+            self.connection._op_commit(self.trans_handle)
+            (h, oid, buf) = self.connection._op_response()
+            delattr(self, "trans_handle")
+            self.connection._transactions.remove(self)
+
+    def rollback(self, retaining=False):
+        if retaining:
+            self.connection._op_rollback_retaining(self.trans_handle)
+            (h, oid, buf) = self.connection._op_response()
+        else:
+            self.connection._op_rollback()
+            (h, oid, buf) = self.connection._op_response(self.trans_handle)
+            delattr(self, "trans_handle")
+            self.connection._transactions.remove(self)
+
+    def close(self):
+        if self.closed:
+            return
+        self.connection._op_rollback(self.trans_handle)
+        (h, oid, buf) = self.connection._op_response()
+        delattr(self, "trans_handle")
+        self.connection._transactions.remove(self)
+
+    @property
+    def connection(self):
+        return self._connection
+
+    @property
+    def closed(self):
+        return not hasattr(self, "trans_handle")
+
